import os
import threading
import time
import webbrowser
import json
from wsgiref.simple_server import make_server
from urllib.parse import parse_qs
from attr import dataclass
from authlib.integrations.requests_client import OAuth2Session
from dotenv import load_dotenv

from common_utils.logger import create_logger
from common_utils.config import ROOT_DIR


@dataclass
class OAuth2Data:
    client_id_env: str
    client_secret_env: str
    redirect_uri_env: str
    scope_env: str
    authorize_url: str
    token_url: str
    state: str | None = None


class OAuth2Handler:
    log = create_logger("OAuth2Handler")
    auth_code = None

    def __init__(self, data: OAuth2Data, cache_path: str | None = None):
        self.client_id = os.environ[data.client_id_env]
        self.client_secret = os.environ[data.client_secret_env]
        self.redirect_uri = os.environ[data.redirect_uri_env]
        self.scope = os.environ[data.scope_env]
        self.authorize_url = data.authorize_url
        self.token_url = data.token_url
        self.state = data.state
        self.port = int(self.redirect_uri.split(":")[-1].split("/")[0])
        self.cache_path = cache_path
        if not cache_path:
            self.cache_name = self.authorize_url.split("/")[2].replace(".", "_")
            self.cache_path = f"{ROOT_DIR}/oauth_tokens/{self.cache_name}.json".replace("\\", "/")

    def callback_app(self, environ, start_response):
        query = environ.get("QUERY_STRING", "")
        params = parse_qs(query)
        self.auth_code = params.get("code", [None])[0]
        start_response("200 OK", [("Content-Type", "text/html")])
        return [b"<h1>Authorization code received. You may close this window.</h1>"]

    def start_temp_server(self):
        httpd = make_server("", self.port, self.callback_app)
        httpd.handle_request()

    def save_token(self, token):
        path = "/".join(self.cache_path.split("/")[:-1])
        os.makedirs(path, exist_ok=True)
        with open(self.cache_path, "w") as file:
            token_str = json.dumps(token)
            file.write(token_str)

    def load_token(self):
        with open(self.cache_path, "r") as file:
            token_str = file.read()
            return json.loads(token_str)

    def get_token(self):
        try:
            token = self.load_token()
            self.log.debug(f"Loaded token from cache: {token}")
            return token
        except FileNotFoundError:
            self.log.debug("No token found in cache. Requesting new token.")

        session = OAuth2Session(
            client_id=self.client_id,
            client_secret=self.client_secret,
            redirect_uri=self.redirect_uri,
            scope=self.scope,
        )
        authorization_url, state = session.create_authorization_url(self.authorize_url)
        server_thread = threading.Thread(target=self.start_temp_server)
        server_thread.daemon = True
        server_thread.start()
        self.log.debug(f"Opening browser to: {authorization_url}")
        webbrowser.open(authorization_url)
        while self.auth_code is None:
            time.sleep(0.1)
        self.log.debug(f"Received auth code: {self.auth_code}")
        token = session.fetch_token(self.token_url, code=self.auth_code)
        self.log.debug(f"Access token data: {token}")
        self.save_token(token)

        return token


if __name__ == "__main__":
    load_dotenv()
    oauth_data = OAuth2Data(
        client_id_env="TICKTICK_CLIENT_ID",
        client_secret_env="TICKTICK_CLIENT_SECRET",
        redirect_uri_env="TICKTICK_REDIRECT_URI",
        scope_env="TICKTICK_SCOPE",
        authorize_url="https://ticktick.com/oauth/authorize",
        token_url="https://ticktick.com/oauth/token",
    )

    oauth_handler = OAuth2Handler(oauth_data)
    oauth_handler.get_token()




import requests
import json
from datetime import datetime, timezone, timedelta
from common_utils.logger import create_logger
from common_utils.web.cookies_handler import CookiesManager, LoginData, LoginSelectors


# TODO: dataclass for habit metadata
# TODO: dataclass for habit checkin

class TicktickHabitHandler:
    """Class that accesses the TickTick habits API. Used to post habit checkins.

    Definitions:
    - Habit: A habit is a task that can be checked in multiple times a day, with a goal value.
             Habits have a unique name and id, stored in their metadata.
    - Checkin: A checkin is a single entry of a habit, with a status and a value.
    """

    status_codes = {0: "Not completed", 1: "Failed", 2: "Completed"}
    urls = {
        "habit": "https://api.ticktick.com/api/v2/habits",
        "query_checkin": "https://api.ticktick.com/api/v2/habitCheckins/query",
        "batch_checkin": "https://api.ticktick.com/api/v2/habitCheckins/batch",
    }
    cookies_manager = CookiesManager(
        login_data=LoginData(
            username_env="TICKTICK_EMAIL",
            password_env="TICKTICK_PASSWORD",
            sign_in_url="https://www.ticktick.com/signin",
            selectors=LoginSelectors(
                username='input[placeholder="Email"]',
                password='#password',
                login_button='#app div[class^=body] button'
            )
        ),
        test_cookies_url="https://api.ticktick.com/api/v2/habits",
        test_cookies_response_fn=lambda data: "errorCode" not in data,
        min_num_cookies=5
    )

    def __init__(self, cookies_path: str | None = None):
        self.log = create_logger("Ticktick Habits")
        if cookies_path:
            self.cookies_manager.cookies_path = cookies_path
        self.headers = self.cookies_manager.get_headers_with_cookies()
        self.habits, self.habit_ids = self._get_all_habits_metadata()

    def _make_api_post_request(self, url_name, payload) -> dict[str, str] | None:
        """Returns response or None if an error occurs"""
        try:
            response = requests.post(url=self.urls[url_name],
                                     data=json.dumps(payload),
                                     headers=self.headers
                                     ).json()
            return response
        except Exception as e:
            self.log.error(f"Error sending API request: {e}")
            return None

    def _get_all_habits_metadata(self) -> tuple[dict, dict]:
        """Get the metadata of all habits and their ids"""
        habit_data = requests.get(self.urls["habit"], headers=self.headers).json()
        if "errorId" in habit_data:
            error_message = f"Error loading habits: {habit_data}"
            self.log.error(error_message)
            raise ValueError(error_message)
        habits = {habit["id"]: habit for habit in habit_data}
        habit_ids = {habit["name"]: habit["id"] for habit in habit_data}
        return habits, habit_ids

    def _post_habit_metadata(self, habit_id):
        # todo: update a single habit metadata, to update checkins quickly
        raise NotImplementedError

    def _get_current_utc_time(self) -> str:
        return datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S") + ".000+0000"

    def _aggregate_single_checkin_data(
        self, habit_name: str, date_stamp: str, status: int | None = None, value: int | None = None
    ) -> dict[str, str | int]:
        """Collect all data needed for a single habit checkin."""
        assert not status and value, "You can only provide status or value, not both"
        assert status is not None or value is not None, "You need to provide either status or value"
        assert habit_name in self.habit_ids, f"Habit {habit_name} not found in habits"

        habit_id = self.habit_ids[habit_name]
        habit_goal = int(self.habits[habit_id]["goal"])
        if not value:
            value = habit_goal if status == 2 else 0
        if not status:
            status = 2 if value >= habit_goal else 0

        return {
            "checkinStamp": date_stamp,
            "checkinTime": self._get_current_utc_time(),
            "goal": habit_goal,
            "habitId": habit_id,
            "opTime": self._get_current_utc_time(),
            "status": status,
            "value": value,
        }

    def post_checkin(self,
                     habit_name: str,
                     date_stamp: str,
                     status: int | None = None,
                     value: int | None = None
                     ) -> None:
        """Post a single habit checkin to the TickTick API.

        Args:
            habit_name: Name of the habit to check-in
            date_stamp: Date of the check-in, in the format YYYYMMDD
            status: Status of the check-in. 0: Not completed, 1: Failed, 2: Completed
            value: The value amount to check in. for habits who require multiple units
        """
        checkin_data = self._aggregate_single_checkin_data(habit_name, date_stamp, status, value)
        self.log.info(
            f"Checking {habit_name} on {date_stamp} as {status}: {value}/{checkin_data['habit_goal']}"
        )
        # create payload depending on if a checkin for that day already exists
        existing_checkin_entry = self.get_checkin(checkin_data["habit_id"], date_stamp)
        payload: dict[str, list[dict[str, str | int]]] = {"add": [], "update": [], "delete": []}
        if existing_checkin_entry:
            checkin_data["id"] = existing_checkin_entry["id"]
            payload["update"].append(checkin_data)
        else:
            payload["add"].append(checkin_data)
        response = self._make_api_post_request("batch_checkin", payload)
        if response:
            self.log.info(f"Checkin for {habit_name} successful with response: {response}")

    def get_checkin(self, habit_id, date_stamp):
        """Retrieve a single checkin entry for a habit on a specific date, or None if not found"""
        date = datetime.strptime(date_stamp, "%Y%m%d")
        after_stamp = (date - timedelta(days=1)).strftime("%Y%m%d")
        payload = {"habitIds": [habit_id], "afterStamp": after_stamp}
        response = self._make_api_post_request(self.urls["checkin"], payload)
        assert response, "No response from API for get_checkin"
        habit_entries = response["checkins"][habit_id]
        for entry in habit_entries:
            if entry["checkinStamp"] == int(date_stamp):
                return entry
        return None

    def get_all_checkins(self,
                         after_stamp: str = "19700101",
                         habit_names: list[str] | str | None = None
                         ) -> list[dict[str, str]]:
        """Get all checkins of all habits (or those provided), after a specific date stamp."""
        assert len(after_stamp) == 8, "Date stamp should be in the format YYYYMMDD"
        if not habit_names:
            habits_ids = list(self.habit_ids.values())
        else:
            habit_names = [habit_names] if isinstance(habit_names, str) else habit_names
            habits_ids = [self.habit_ids[habit] for habit in habit_names]
        payload = {"habitIds": habits_ids, "afterStamp": after_stamp}
        response = self._make_api_post_request("query_checkin", payload)
        assert response, "No response from API"
        habit_entries = response["checkins"]
        flattened_habit_entries_list = [
            {**entry, "habitName": self.habits[habit_id]["name"]}
            for habit_id, entries in habit_entries.items()
            for entry in entries
        ]
        return flattened_habit_entries_list


if __name__ == "__main__":
    pass



import yaml
from dotenv import load_dotenv
import os
from os import getenv as secret  # noqa: F401

from common_utils.logger import create_logger

logged_root_dirs = []


def get_root_dir(dunder_file: str) -> str | None:
    """ Returns the root directory of the project, called using __file__.

    Uses either 'src', 'venv' or DOCKER_WORKDIR env-var to determine the root directory.
    """
    root_dir = None
    try:
        if "venv" in os.path.abspath(dunder_file):
            root_dir = os.path.abspath(dunder_file).split("venv")[0][:-1]
        elif "src" in os.path.abspath(dunder_file):
            root_dir = os.path.abspath(dunder_file).split("src")[0][:-1]
        elif os.getenv('DOCKER_WORKDIR'):
            root_dir = os.getenv('DOCKER_WORKDIR')
        else:
            if dunder_file != __file__:
                logger.error(f"Cannot get root dir, as {dunder_file} not in src or venv dir")
            else:
                logger.debug(f"Cannot get ROOT_DIR, as {dunder_file} not in src or venv dir")
    except Exception:
        logger.error(f"Could not get root dir from {dunder_file}")
    finally:
        # remove /. end of root_dir
        if root_dir and root_dir[-2:] == '\\.':
            root_dir = root_dir[:-2]
        elif root_dir and root_dir[-1] == '\\':
            root_dir = root_dir[:-1]
        if os.getenv('DONT_PRINT_ROOT_DIR'):
            return root_dir
        logger.debug(f"{'ROOT_DIR' if dunder_file == __file__ else 'get_root_dir'}: {root_dir}"
                     f"  -  {os.listdir(root_dir) if root_dir not in logged_root_dirs else ''}")
        logged_root_dirs.append(root_dir)
        return root_dir


def load_config_yaml(file_path: str) -> dict | None:
    """ Loads a yaml file from the given path and returns the data as a dict. """
    try:
        _stream = open(file_path, "r", encoding="utf-8")
        return yaml.safe_load(_stream)
    except Exception as e:
        logger.error(f"Could not load config from {file_path}")
        return None


def config_entry(key):
    return CONFIG[key]


# Load default config from config.yml using ROOT_DIR
logger = create_logger("Config Helper")
load_dotenv()
ROOT_DIR = get_root_dir(__file__)
if ROOT_DIR:
    CONFIG = load_config_yaml(f"{ROOT_DIR}/config.yml")
else:
    logger.debug("ROOT_DIR is None, trying to load CONFIG from config.yml in current dir")
    CONFIG = load_config_yaml("config.yml")



import requests
import json
import os

from common_utils.logger import create_logger


class FirebaseClient:
    log = create_logger("FirebaseClient")
    headers = {"Content-Type": "application/x-www-form-urlencoded"}

    def __init__(self, realtime_db_url: str | None = None):
        """Firebase client to interact with a realtime database

        Currently only supports unauthenticated access."""
        self.database_url = realtime_db_url or os.getenv("FIREBASE_REALTIME_DB_URL")
        if not self.database_url:
            raise ValueError("Could not load firebase project from environment variables")

    def get_list(self, ref: str, max_results: int = 100, convert_to_list: bool = True):
        """Get a list of child-entries from an object in firebase."""
        params = {"orderBy": '"$key"', "limitToFirst": str(max_results)}
        reference_url = f"{self.database_url}/{ref}.json"
        try:
            response = requests.get(url=reference_url, params=params)
            self.log.debug(f"Getting List: {response.text}")
            if convert_to_list:
                list_json = response.json()
                return [list_json[key] for key in list_json]
            else:
                return response.json()
        except Exception as e:
            self.log.error(f"Failed to get list from {ref}: {e}")
            return None

    def delete_entry(self, ref: str):
        """Delete an object entry from firebase"""
        try:
            requests.delete(url=f"{self.database_url}/{ref}.json")
            self.log.debug(f"Deleted entry {ref} from firebase")
        except Exception as e:
            self.log.error(f"Failed to delete entry {ref}: {e}")

    def get_entry(self, ref: str):
        """Get an object entry from firebase"""
        try:
            response = requests.get(url=f"{self.database_url}/{ref}.json").json()
            self.log.debug(f"Got entry {ref} from firebase: {response}")
            return response
        except Exception as e:
            self.log.error(f"Failed to get entry {ref}: {e}")
            return None

    def set_entry(self, ref: str, data: dict):
        """Set an object entry in firebase"""
        try:
            url = f"{self.database_url}/{ref}.json"
            response = requests.put(url=url, headers=self.headers, data=json.dumps(data))
            self.log.debug(f"Set entry {ref} with data {data} in firebase: {response.text}")
        except Exception as e:
            self.log.error(f"Failed to set entry {ref}: {e}")


if __name__ == "__main__":
    firebase = FirebaseClient()



from notion_client import Client
import pandas as pd
from concurrent.futures import ThreadPoolExecutor

from common_utils.config import secret, load_config_yaml
from common_utils.logger import create_logger


class NotionClient:
    """Wrapper for the Notion API. Loads data from a database.

    Attributes:
        api: NotionClient object
        projects_db: List of all projects from the projects database (config)
        project_names: dict, project ids as keys, project names as values
    """

    def __init__(self, notion_secret=None):
        if not notion_secret:
            notion_secret = secret("NOTION_SECRET")
        assert notion_secret, "Could not load NOTION_SECRET from variable or .env"
        self.log = create_logger("Notion")
        self.api = Client(auth=secret("NOTION_SECRET"))

    def _get_title(self, page, property_name) -> str:
        return page["properties"][property_name]["title"][0]["plain_text"]

    def _get_emoji(self, page) -> str:
        return page["icon"]["emoji"]

    def _get_plaintext(self, page, property_name) -> str:
        plain_texts = [item["plain_text"] for item in page["properties"][property_name]["rich_text"]]
        return "".join(plain_texts)

    def _get_number(self, page, property_name) -> float:
        return page["properties"][property_name]["number"]

    def _get_select(self, page, property_name) -> object:
        select = page["properties"][property_name]["select"]
        return select

    def _get_select_name(self, page, property_name) -> str:
        select = page["properties"][property_name]["select"]["name"]
        return select

    def _get_multi_select(self, page, property_name) -> list[object]:
        multiselect = page["properties"][property_name]["multi_select"]
        return multiselect

    def _get_multi_select_names(self, page, property_name) -> list[str]:
        multiselect = page["properties"][property_name]["multi_select"]
        return [item["name"] for item in multiselect]

    def _get_date(self, task, property_name) -> tuple[str, str]:
        """Returns the start and end dates as strings of a task, empty if a date is not set"""
        date_property = task["properties"][property_name]["date"]
        start_date, end_date = None, None
        if date_property is not None and date_property["start"] is not None:
            start_date = date_property["start"]
        if date_property is not None and date_property["end"] is not None:
            end_date = date_property["end"]
        return start_date, end_date

    def get_properties(self, page, property_mapping) -> dict:
        """ Returns all properties of a page using the property_mapping """
        properties = {"id": page["id"]}
        for name, property_type in property_mapping.items():
            try:
                match property_type:
                    case "title": properties[name] = self._get_title(page, name)
                    case "emoji": properties[name] = self._get_emoji(page)
                    case "content_simple": properties[name] = self.get_page_content(page["id"], "simple")
                    case "content_markdown": properties[name] = self.get_page_content(page["id"], "markdown")
                    case "text": properties[name] = self._get_plaintext(page, name)
                    case "number": properties[name] = self._get_number(page, name)
                    case "select": properties[name] = self._get_select(page, name)
                    case "select_name": properties[name] = self._get_select_name(page, name)
                    case "multi_select": properties[name] = self._get_multi_select(page, name)
                    case "multi_select_names": properties[name] = self._get_multi_select_names(page, name)
                    case "date":
                        start_date, end_date = self._get_date(page, name)
                        properties[name] = start_date
                        properties[name + "_end"] = end_date
                    case _: properties[name] = page["properties"][property_type]
            except Exception as e:
                properties[name] = None
        return properties

    def get_database(self, database_id: str, properties_mapping: dict[str, str], filter={}, sort={}) -> pd.DataFrame:
        """Returns dataframe with data from a specific database.

        Args:
            database_id: Notion ID of the database
            properties_mapping: Mapping of database properties to their types.
                Example: {"Name": "title", "Status": "select"}
                Types: title, emoji, text, number, select, select_name,
                       multi_select, multi_select_names, date
            filter: Filter for the database query
            sort: Sorting for the database query

        Returns:
            pd.DataFrame: Dataframe with all tasks and their properties
        """

        all_entries = pd.DataFrame()
        query_parameter = {}
        if filter:
            query_parameter["filter"] = filter
        if sort:
            query_parameter["sorts"] = sort
        query = self.api.databases.query(database_id, **query_parameter)["results"]
        for page in query:
            properties = self.get_properties(page, properties_mapping)
            all_entries = all_entries._append(properties, ignore_index=True)
        return all_entries

    def get_multiple_databases(self, db_config_dict: dict[dict]) -> dict[str, pd.DataFrame]:
        with ThreadPoolExecutor(max_workers=2) as executor:
            thread_list = {}
            results = {}
            for name, db_config in db_config_dict.items():
                thread = executor.submit(self.get_database, **db_config)
                thread_list[name] = thread
            for name, thread in thread_list.items():
                results[name] = thread.result()
        return results

    def _get_text_part_plaintext(self, text_part: dict) -> str:
        if 'text' in text_part:
            return text_part['text']['content']
        elif 'mention' in text_part:
            if 'link_mention' in text_part['mention']:
                return text_part['mention']['link_mention']['href']
        else:
            return ""

    def get_page_content(self, page_id: str, parse_mode: str) -> str:
        assert parse_mode in ['markdown', 'simple'], "parse_mode must be 'markdown' or 'simple'"
        blocks = self.api.blocks.children.list(block_id=page_id)
        numbered_list_count = 1
        entire_text = ""
        for block in blocks['results']:
            try:
                if numbered_list_count > 1 and block['type'] != 'numbered_list_item':
                    numbered_list_count = 1
                match block['type']:
                    case 'paragraph':
                        text = ''.join([self._get_text_part_plaintext(text_part) for text_part in
                                        block['paragraph']['rich_text']])
                    case 'heading_1' | 'heading_2' | 'heading_3':
                        text = ''.join([text_part['text']['content'] for text_part in
                                        block[block['type']]['rich_text']])
                        heading_level = int(block['type'].split('_')[1])
                        text = f"{'#' * heading_level} {text}" if parse_mode == 'markdown' else f"*{text}*"
                    case 'bookmark':
                        url = block['bookmark']['url']
                        text = f"{url}"
                    case 'bulleted_list_item':
                        text = ''.join([self._get_text_part_plaintext(text_part) for text_part in
                                        block['bulleted_list_item']['rich_text']])
                        text = f"- {text}" if parse_mode == 'markdown' else f"• {text}"
                    case 'numbered_list_item':
                        text = ''.join([self._get_text_part_plaintext(text_part) for text_part in
                                        block['numbered_list_item']['rich_text']])
                        text = f"{numbered_list_count}. {text}"
                        numbered_list_count += 1
                    case 'quote':
                        text = ''.join(
                            [self._get_text_part_plaintext(text_part) for text_part in
                             block['quote']['rich_text']])
                        text = f"> {text}" if parse_mode == 'markdown' else f"\"**{text}**\""
                    case 'code':
                        text = block['code']['text']
                        text = f"```{text}```"
                    case 'to_do':
                        text = block['to_do']['text']
                        text = f"[ ] {text}"
                    case 'toggle':
                        text = block['toggle']['text']
                    case 'child_page':
                        text = f"_{block['child_page']['title']}_"
                    case 'divider':
                        text = "---" if parse_mode == 'markdown' else ""
                    case 'callout':
                        text = block['callout']['text']
                        text = f"**{text}**"
                    case _:
                        text = ""
                entire_text += f"{text}\n"
            except Exception as e:
                print(e)
        if parse_mode == 'simple':
            entire_text = entire_text.replace("https://", "")
        return entire_text

    def add_property(self, prop_type, value):
        match prop_type:
            case 'title': return {"title": [{"text": {"content": value}}]}
            case 'number': return {"number": value}
            case 'text': return {"rich_text": [{"text": {"content": value}}]}
            case 'date': return {"date": {"start": value}}
            case 'select': return {"select": {"name": value}}
            case 'multi_select': return {"multi_select": [{"name": item} for item in value]}
            case 'checkbox': return {"checkbox": value}
            case 'url': return {"url": value}
            case 'email': return {"email": value}
            case 'phone_number': return {"phone_number": value}
            case 'formula': return {"formula": {"expression": value}}
            case 'relation': return {"relation": value}
            case 'rollup': return {"rollup": {"relation": value}}
            case _: raise ValueError(f"Unsupported property type: {prop_type}")

    def create_db_entry(self, database_id, properties=None, data=None, properties_mapping=None):
        """Creates a new entry in a database.

        Currently only supports title, text, number, select, multi_select, date, url, email,
        phone_number, formula, relation, rollup, checkbox

        Args:
            database_id: Notion ID of the database
            properties: Dict with properties and their values
                Example: {"Name": {"type": "title", "value": "My task"},
                          "Status": {"type": "select", "value": "In Progress"}}
            """
        assert properties or (data and properties_mapping), "Either properties or data and properties_mapping must be set"
        if data and properties_mapping:
            properties = self._convert_data_to_notion_properties(data, properties_mapping)
        notion_properties = {}
        for key, prop in properties.items():
            notion_properties[key] = self.add_property(prop['type'], prop['value'])
        try:
            self.api.pages.create(parent={"database_id": database_id}, properties=notion_properties)
        except Exception as e:
            print("An error occurred:", e)

    @staticmethod
    def _convert_data_to_notion_properties(data: dict | pd.Series, properties_mapping: dict) -> dict:
        # TODO: integrate into my-common-utils, automatically detect type from notion database schema
        notion_properties = {}
        for key, value in data.items():
            try:
                entry_type = properties_mapping[key]
                try:
                    value = float(value)
                except ValueError:
                    pass
                notion_properties[key] = {'type': entry_type, 'value': value}
            except KeyError:
                pass
        return notion_properties


if __name__ == "__main__":
    from dotenv import load_dotenv
    from common_utils.config import load_config_yaml
    load_dotenv()
    notion = NotionClient()
    config = load_config_yaml("config.yml")
    db_config_dict = config["NOTION_DATABASES"]
    fitness_creation_dict = db_config_dict["fitness"]
    fitness_df = notion.create_db_entry(**fitness_creation_dict)




    print()



from dataclasses import dataclass
from typing import Literal
from time import sleep
from selenium.webdriver.common.by import By
from selenium import webdriver
import undetected_chromedriver as uc

from common_utils.logger import create_logger



@dataclass
class DriverAction:
    action: Literal["url", "sleep", "click", "send_keys", "get_text", "get_texts", "get_class", "get_classes"]
    identifier: str | None = None
    input: str | int | None = None
    key: str | None = None



class SeleniumHandler:
    log = create_logger("SeleniumHandler")

    def __init__(self,
                 headless: bool = False,
                 undetected: bool = False,
                 verbose: bool = True,
                 raise_exceptions: bool = True,
                 download_driver: bool = False
    ) -> None:
        self.headless = headless
        self.undetected = undetected
        self.verbose = verbose
        self.raise_exceptions = raise_exceptions
        self.download_driver = download_driver

    def log_msg(self, message: str, level: Literal["debug", "info", "warning", "error"] = "info"):
        if self.verbose is False:
            return
        match level:
            case "debug":
                self.log.debug(message)
            case "info":
                self.log.info(message)
            case "warning":
                self.log.warning(message)
            case "error":
                self.log.error(message)
            case _:
                self.log.info(message)

    def get_driver(self):
        kwargs = {}
        if self.download_driver and not self.undetected:
            from selenium.webdriver.chrome.service import Service
            from webdriver_manager.chrome import ChromeDriverManager
            kwargs['service'] = Service(ChromeDriverManager().install())

        if self.headless and self.undetected:
            options = uc.ChromeOptions()
            options.add_argument("--headless")
            kwargs['options'] = options
        elif self.headless:
            options = webdriver.ChromeOptions()
            options.add_argument("--headless")
            options.add_argument("--disable-gpu")
            options.add_argument("--no-sandbox")
            options.add_argument("--disable-dev-shm-usage")
            kwargs['options'] = options

        if self.undetected:
            return uc.Chrome(**kwargs)
        else:
            return webdriver.Chrome(**kwargs)

    def run_actions(self, actions: list[DriverAction], driver: webdriver.Chrome | None = None):
        _driver = driver if driver is not None else self.get_driver()
        results = {}
        for action in actions:
            try:
                result = self.run_action(action, _driver)
                if action.key is not None:
                    assert action.key not in results, f"Result key already exists: {action.key}"
                    results[action.key] = result
            except Exception as e:
                self.log_msg(f"Error in action: {action} - {e}", level='error')
                if self.raise_exceptions:
                    raise e
        if driver is None:
            _driver.quit()
        return results

    def run_action(self, action: DriverAction, driver: webdriver.Chrome):
        result = None
        match action.action:
            case "url":
                driver.get(action.identifier)
                self.log_msg(f"Opened URL: {action.identifier}", level='debug')
            case "send_keys":
                driver.find_element(By.CSS_SELECTOR, action.identifier).send_keys(action.input)
                self.log_msg(f"Sent keys: {action.input} to {action.identifier}", level='debug')
            case "click":
                driver.find_element(By.CSS_SELECTOR, action.identifier).click()
                self.log_msg(f"Clicked on: {action.identifier}", level='debug')
            case "sleep":
                self.log_msg(f"Sleeping for: {action.input} seconds", level='debug')
                sleep(action.input)
            case "get_text":
                result = driver.find_element(By.CSS_SELECTOR, action.identifier).text
                self.log_msg(f"Storing text: {result}", level='debug')
            case "get_texts":
                result = [element.text for element in driver.find_elements(By.CSS_SELECTOR, action.identifier)]
                self.log_msg(f"Storing texts: {result}", level='debug')
            case "get_class":
                result = driver.find_element(By.CSS_SELECTOR, action.identifier).get_attribute("class")
                self.log_msg(f"Storing class: {result}", level='debug')
            case "get_classes":
                result = [element.get_attribute("class") for element in driver.find_elements(By.CSS_SELECTOR, action.identifier)]
                self.log_msg(f"Storing classes: {result}", level='debug')
            case _:
                self.log_msg(f"Invalid action: {action.action}", level='error')
                if self.raise_exceptions:
                    raise ValueError(f"Invalid action: {action.action}")
        return result


if __name__ == '__main__':
    ud = SeleniumHandler(undetected=True).get_driver()
    ud.get("https://www.google.com")
    sleep(2)

    ud2 = SeleniumHandler(undetected=True, headless=True).get_driver()
    ud2.get("https://www.google.com")
    google_text = ud2.find_element(By.CSS_SELECTOR, 'body').text
    print(google_text)